"""Type definitions for Claude SDK.

This package re-exports all types for backwards compatibility.
Types are organized into submodules:
- base: Type aliases, literals, and thinking configuration
- agents: Agent definitions and presets
- permissions: Permission system types
- hooks: Hook system types (inputs, outputs, matchers, callbacks)
- mcp: MCP server and plugin configuration
- sandbox: Sandbox configuration
- messages: Content blocks, message types, and stream events
- options: ClaudeAgentOptions
- control: SDK control protocol types
"""

from __future__ import annotations

from .agents import AgentDefinition, SystemPromptPreset, ToolsPreset
from .base import (
    ApiKeySource,
    PermissionMode,
    PermissionBehavior,
    ReasoningEffort,
    SdkBeta,
    SettingSource,
    StopReason,
    ThinkingConfig,
    ThinkingConfigAdaptive,
    ThinkingConfigDisabled,
    ThinkingConfigEnabled,
)
from .control import (
    ControlErrorResponse,
    ControlRequestUnion,
    ControlResponse,
    SDKControlInitializeRequest,
    SDKControlInterruptRequest,
    SDKControlMcpMessageRequest,
    SDKControlPermissionRequest,
    SDKControlRequest,
    SDKControlResponse,
    SDKControlRewindFilesRequest,
    SDKControlSetPermissionModeRequest,
    SDKControlStopTaskRequest,
    SDKHookCallbackRequest,
    control_request_adapter,
)
from .hooks import (
    AsyncHookJSONOutput,
    BaseHookInput,
    HookCallback,
    HookContext,
    HookEvent,
    HookInput,
    HookJSONOutput,
    HookMatcher,
    HookSpecificOutput,
    NotificationHookInput,
    NotificationHookSpecificOutput,
    PermissionRequestHookInput,
    SessionEndHookInput,
    SessionStartHookInput,
    SetupHookInput,
    PermissionRequestHookSpecificOutput,
    PostToolUseFailureHookInput,
    PostToolUseFailureHookSpecificOutput,
    PostToolUseHookInput,
    PostToolUseHookSpecificOutput,
    PreCompactHookInput,
    PreToolUseHookInput,
    PreToolUseHookSpecificOutput,
    SessionStartHookSpecificOutput,
    StopHookInput,
    SubagentStartHookInput,
    TaskCompletedHookInput,
    TeammateIdleHookInput,
    SubagentStartHookSpecificOutput,
    SubagentStopHookInput,
    SyncHookJSONOutput,
    UserPromptSubmitHookInput,
    UserPromptSubmitHookSpecificOutput,
    WorktreeCreateHookInput,
    WorktreeRemoveHookInput,
)
from .input_types import (
    TOOL_INPUT_TYPES,
    AgentInput,
    AskUserQuestion,
    AskUserQuestionInput,
    AskUserQuestionOption,
    BashInput,
    BashOutputInput,
    ExitPlanModeInput,
    FileEditInput,
    FileReadInput,
    FileWriteInput,
    GlobInput,
    GrepInput,
    KillShellInput,
    ListMcpResourcesInput,
    NotebookEditInput,
    ReadMcpResourceInput,
    TodoItem,
    TodoWriteInput,
    ToolInput,
    WebFetchInput,
    WebSearchInput,
)
from .mcp import (
    JSONRPC_VERSION,
    JSONRPCError,
    JSONRPCErrorResponse,
    JSONRPCMessage,
    JSONRPCNotification,
    JSONRPCRequest,
    JSONRPCResponse,
    JSONRPCResultResponse,
    McpConnectionStatus,
    McpHttpServerConfig,
    McpSdkServerConfig,
    McpServerConfig,
    McpServerStatusEntry,
    McpServerVersionInfo,
    McpSSEServerConfig,
    McpStatusResponse,
    McpStdioServerConfig,
    McpToolStatus,
    RequestId,
    SdkPluginConfig,
)
from .output_types import (
    TOOL_USE_RESULT_TYPES,
    AgentAsyncLaunchedOutput,
    AgentCacheCreation,
    AgentCompletedOutput,
    AgentOutput,
    AgentOutputTextContent,
    AgentOutputUsage,
    AgentServerToolUse,
    AgentSubAgentEnteredOutput,
    AskUserQuestionItem,
    AskUserQuestionOutput,
    BashOutput,
    BashToolUseResult,
    ConfigOutput,
    EditOutput,
    EditToolUseResult,
    EnterWorktreeOutput,
    ExitPlanModeOutput,
    GitDiff,
    GlobOutput,
    GrepOutput,
    ListMcpResourcesOutput,
    McpOutput,
    McpResourceEntry,
    NotebookEditOutput,
    ReadFileInfo,
    ReadImageDimensions,
    ReadImageFileInfo,
    ReadImageOutput,
    ReadMcpResourceContentItem,
    ReadMcpResourceOutput,
    ReadNotebookFileInfo,
    ReadNotebookOutput,
    ReadOutput,
    ReadPartsFileInfo,
    ReadPartsOutput,
    ReadPdfFileInfo,
    ReadPdfOutput,
    ReadTextOutput,
    ReadToolUseResult,
    StructuredPatchHunk,
    SubscribeMcpResourceOutput,
    SubscribePollingOutput,
    TaskInfo,
    TaskOutputResult,
    TaskStopOutput,
    TodoUseResultItem,
    TodoWriteOutput,
    TodoWriteToolUseResult,
    ToolUseResult,
    UnsubscribeMcpResourceOutput,
    UnsubscribePollingOutput,
    WebFetchOutput,
    WebSearchHit,
    WebSearchOutput,
    WebSearchToolResult,
    WriteOutput,
    WriteToolUseResult,
)
from .content_blocks import (
    ContentBlock,
    TextBlock,
    ThinkingBlock,
    ToolResultBlock,
    ToolUseBlock,
    content_block_adapter,
)

from .messages import (
    AssistantMessage,
    AssistantMessageError,
    BaseSystemMessage,
    CompactBoundarySystemMessage,
    HookResponseSystemMessage,
    HookStartedSystemMessage,
    InitSystemMessage,
    McpServerStatus,
    Message,
    ModelUsage,
    BaseResultMessage,
    ResultErrorMessage,
    ResultMessage,
    ResultSuccessMessage,
    SDKPermissionDenial,
    SDKSessionInfo,
    StatusSystemMessage,
    StreamEvent,
    AuthStatusMessage,
    FilesPersistedSystemMessage,
    HookProgressSystemMessage,
    TaskNotificationSystemMessage,
    TaskProgressSystemMessage,
    TaskProgressUsage,
    TaskStartedSystemMessage,
    ToolProgressMessage,
    ToolUseSummaryMessage,
    TriggerMetadata,
    Usage,
    UserMessage,
    UserPromptMessage,
    UserPromptMessageContent,
    system_message_adapter,
)
from .options import ClaudeAgentOptions
from .options import (
    BaseSessionConfig,
    ContinueLatest,
    ListSessionsOptions,
    NewSession,
    ResumeSession,
    SessionConfig,
    resolve_session_config,
)
from .permissions import (
    CanUseTool,
    OnUserQuestion,
    PermissionResult,
    PermissionResultAllow,
    PermissionResultDeny,
    PermissionRuleValue,
    PermissionUpdate,
    PermissionUpdateDestination,
    ToolPermissionContext,
)
from .sandbox import SandboxIgnoreViolations, SandboxNetworkConfig, SandboxSettings

__all__ = [
    # base
    "ApiKeySource",
    "PermissionMode",
    "ReasoningEffort",
    "SdkBeta",
    "SettingSource",
    "StopReason",
    "ThinkingConfig",
    "ThinkingConfigAdaptive",
    "ThinkingConfigDisabled",
    "ThinkingConfigEnabled",
    # agents
    "AgentDefinition",
    "SystemPromptPreset",
    "ToolsPreset",
    # permissions
    "CanUseTool",
    "OnUserQuestion",
    "PermissionBehavior",
    "PermissionResult",
    "PermissionResultAllow",
    "PermissionResultDeny",
    "PermissionRuleValue",
    "PermissionUpdate",
    "PermissionUpdateDestination",
    "ToolPermissionContext",
    # hooks
    "AsyncHookJSONOutput",
    "BaseHookInput",
    "HookCallback",
    "HookContext",
    "HookEvent",
    "HookInput",
    "HookJSONOutput",
    "HookMatcher",
    "HookSpecificOutput",
    "NotificationHookInput",
    "NotificationHookSpecificOutput",
    "PermissionRequestHookInput",
    "SessionEndHookInput",
    "SessionStartHookInput",
    "SetupHookInput",
    "PermissionRequestHookSpecificOutput",
    "PostToolUseFailureHookInput",
    "PostToolUseFailureHookSpecificOutput",
    "PostToolUseHookInput",
    "PostToolUseHookSpecificOutput",
    "PreCompactHookInput",
    "PreToolUseHookInput",
    "PreToolUseHookSpecificOutput",
    "SessionStartHookSpecificOutput",
    "StopHookInput",
    "SubagentStartHookInput",
    "TaskCompletedHookInput",
    "TeammateIdleHookInput",
    "SubagentStartHookSpecificOutput",
    "SubagentStopHookInput",
    "SyncHookJSONOutput",
    "UserPromptSubmitHookInput",
    "UserPromptSubmitHookSpecificOutput",
    "WorktreeCreateHookInput",
    "WorktreeRemoveHookInput",
    # mcp
    "JSONRPC_VERSION",
    "JSONRPCError",
    "JSONRPCErrorResponse",
    "JSONRPCMessage",
    "JSONRPCNotification",
    "JSONRPCRequest",
    "JSONRPCResponse",
    "JSONRPCResultResponse",
    "McpConnectionStatus",
    "McpHttpServerConfig",
    "McpSdkServerConfig",
    "McpServerConfig",
    "McpServerStatusEntry",
    "McpServerVersionInfo",
    "McpSSEServerConfig",
    "McpStatusResponse",
    "McpStdioServerConfig",
    "McpToolStatus",
    "RequestId",
    "SdkPluginConfig",
    # sandbox
    "SandboxIgnoreViolations",
    "SandboxNetworkConfig",
    "SandboxSettings",
    # messages
    "AssistantMessage",
    "AssistantMessageError",
    "BaseSystemMessage",
    "CompactBoundarySystemMessage",
    "ContentBlock",
    "HookResponseSystemMessage",
    "HookStartedSystemMessage",
    "InitSystemMessage",
    "McpServerStatus",
    "Message",
    "ModelUsage",
    "BaseResultMessage",
    "ResultErrorMessage",
    "ResultMessage",
    "ResultSuccessMessage",
    "SDKPermissionDenial",
    "SDKSessionInfo",
    "StatusSystemMessage",
    "StreamEvent",
    "AuthStatusMessage",
    "FilesPersistedSystemMessage",
    "HookProgressSystemMessage",
    "TaskNotificationSystemMessage",
    "TaskProgressSystemMessage",
    "TaskProgressUsage",
    "TaskStartedSystemMessage",
    "TextBlock",
    "ThinkingBlock",
    "ToolProgressMessage",
    "ToolResultBlock",
    "ToolUseBlock",
    "ToolUseSummaryMessage",
    "TriggerMetadata",
    "Usage",
    "UserMessage",
    "UserPromptMessage",
    "UserPromptMessageContent",
    "content_block_adapter",
    "system_message_adapter",
    # options
    "ClaudeAgentOptions",
    "BaseSessionConfig",
    "ContinueLatest",
    "ListSessionsOptions",
    "NewSession",
    "ResumeSession",
    "SessionConfig",
    "resolve_session_config",
    # control
    "ControlErrorResponse",
    "ControlRequestUnion",
    "ControlResponse",
    "SDKControlInitializeRequest",
    "SDKControlInterruptRequest",
    "SDKControlMcpMessageRequest",
    "SDKControlPermissionRequest",
    "SDKControlRequest",
    "SDKControlResponse",
    "SDKControlRewindFilesRequest",
    "SDKControlSetPermissionModeRequest",
    "SDKControlStopTaskRequest",
    "SDKHookCallbackRequest",
    "control_request_adapter",
    "ToolInput",
    "TOOL_INPUT_TYPES",
    "AgentInput",
    "AskUserQuestion",
    "AskUserQuestionInput",
    "AskUserQuestionOption",
    "BashInput",
    "BashOutputInput",
    "FileEditInput",
    "FileReadInput",
    "FileWriteInput",
    "GlobInput",
    "GrepInput",
    "KillShellInput",
    "NotebookEditInput",
    "WebFetchInput",
    "WebSearchInput",
    "TodoItem",
    "TodoWriteInput",
    "ExitPlanModeInput",
    "ListMcpResourcesInput",
    "ReadMcpResourceInput",
    # output_types (actual tool_use_result wire format)
    "TOOL_USE_RESULT_TYPES",
    "AgentAsyncLaunchedOutput",
    "AgentCacheCreation",
    "AgentCompletedOutput",
    "AgentOutput",
    "AgentOutputTextContent",
    "AgentOutputUsage",
    "AgentServerToolUse",
    "AgentSubAgentEnteredOutput",
    "AskUserQuestionItem",
    "AskUserQuestionOutput",
    "BashOutput",
    "ConfigOutput",
    "EditOutput",
    "EnterWorktreeOutput",
    "ExitPlanModeOutput",
    "GitDiff",
    "GlobOutput",
    "GrepOutput",
    "ListMcpResourcesOutput",
    "McpOutput",
    "McpResourceEntry",
    "NotebookEditOutput",
    "ReadFileInfo",
    "ReadImageDimensions",
    "ReadImageFileInfo",
    "ReadImageOutput",
    "ReadMcpResourceContentItem",
    "ReadMcpResourceOutput",
    "ReadNotebookFileInfo",
    "ReadNotebookOutput",
    "ReadOutput",
    "ReadPartsFileInfo",
    "ReadPartsOutput",
    "ReadPdfFileInfo",
    "ReadPdfOutput",
    "ReadTextOutput",
    "StructuredPatchHunk",
    "SubscribeMcpResourceOutput",
    "SubscribePollingOutput",
    "TaskInfo",
    "TaskOutputResult",
    "TaskStopOutput",
    "TodoWriteOutput",
    "ToolUseResult",
    "UnsubscribeMcpResourceOutput",
    "UnsubscribePollingOutput",
    "WebFetchOutput",
    "WebSearchHit",
    "WebSearchOutput",
    "WebSearchToolResult",
    "WriteOutput",
    # backwards-compat aliases (old tool_use_results.py names)
    "BashToolUseResult",
    "EditToolUseResult",
    "ReadToolUseResult",
    "TodoUseResultItem",
    "TodoWriteToolUseResult",
    "WriteToolUseResult",
]



"""Base type aliases, literals, and thinking configuration types."""

from __future__ import annotations

import sys
from typing import Literal, TypedDict

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


PermissionMode = Literal[
    "default",  #  Standard behavior, prompts for dangerous operations
    "acceptEdits",  # Auto-accept file edit operations
    "bypassPermissions",  # Bypass all permission checks (requires allowDangerouslySkipPermissions)
    "plan",  # Planning mode, no actual tool execution
    "delegate",  # Delegate mode, restricts to only Teammate and Task tools
    "dontAsk",  # Don't prompt for permissions, deny if not pre-approved
]
SdkBeta = Literal["context-1m-2025-08-07"]  # see https://docs.anthropic.com/en/api/beta-headers
ModelName = Literal["sonnet", "opus", "haiku"]
PermissionBehavior = Literal["allow", "deny", "ask"]
ReasoningEffort = Literal["low", "medium", "high", "max"]
TaskStatus = Literal["completed", "failed", "stopped"]
StopReason = Literal[
    "end_turn",
    "max_tokens",
    "stop_sequence",
    "tool_use",
    "pause_turn",
    "refusal",
    "model_context_window_exceeded",
]
ApiKeySource = Literal["none", "env", "config", "ANTHROPIC_API_KEY"]
SettingSource = Literal["user", "project", "local"]
ToolName = Literal[
    "Task",
    "TaskOutput",
    "Bash",
    "Glob",
    "Grep",
    "ExitPlanMode",
    "Read",
    "Edit",
    "Write",
    "NotebookEdit",
    "WebFetch",
    "TodoWrite",
    "WebSearch",
    "TaskStop",
    "AskUserQuestion",
    "Skill",
    "EnterPlanMode",
    "EnterWorktree",
    "ToolSearch",
]
IS_DEV = "pytest" in sys.modules


class ClaudeCodeBaseModel(BaseModel):
    """Base model for all Claude Code Pydantic models."""

    model_config = ConfigDict(
        populate_by_name=True,
        alias_generator=to_camel,
        extra="forbid" if IS_DEV else "ignore",
    )


# Thinking configuration types
class ThinkingConfigAdaptive(TypedDict):
    """Adaptive thinking configuration - model decides thinking budget."""

    type: Literal["adaptive"]


class ThinkingConfigEnabled(TypedDict):
    """Enabled thinking configuration with explicit token budget."""

    type: Literal["enabled"]
    budget_tokens: int


class ThinkingConfigDisabled(TypedDict):
    """Disabled thinking configuration."""

    type: Literal["disabled"]


ThinkingConfig = ThinkingConfigAdaptive | ThinkingConfigEnabled | ThinkingConfigDisabled



"""End-to-end tests for MCP tools over the wire.

Tests whether content blocks from an external MCP server
(FastMCP stdio) are correctly received through the Claude Code CLI.
"""

from __future__ import annotations

from pathlib import Path
import sys
from typing import TYPE_CHECKING, Any

import pytest

from clawd_code_sdk import (
    AssistantMessage,
    ClaudeAgentOptions,
    ClaudeSDKClient,
    ResultMessage,
    UserMessage,
)
from clawd_code_sdk.models.content_blocks import TextBlock, ToolResultBlock, ToolUseBlock
from clawd_code_sdk.models.messages import ToolProgressMessage


if TYPE_CHECKING:
    from clawd_code_sdk.models.messages import Message


@pytest.mark.e2e
@pytest.mark.asyncio
async def test_mcp_image_tool_wire_format():
    """Test that image content from an MCP tool flows through the wire protocol.

    Configures Claude Code with a FastMCP stdio server that has a single tool
    returning a PNG image, then asks Claude to call it and inspects what
    content blocks come back.
    """
    mcp_server_path = str(Path(__file__).parent.parent / "mcp_server.py")

    options = ClaudeAgentOptions(
        mcp_servers={
            "image_test": {"type": "stdio", "command": sys.executable, "args": [mcp_server_path]},
        },
        permission_mode="bypassPermissions",
        allow_dangerously_skip_permissions=True,
        max_turns=3,
    )

    messages: list[Message] = []
    async with ClaudeSDKClient(options=options) as client:
        await client.query(
            "Call the mcp__image_test__get_test_image tool and describe what you see."
        )
        async for message in client.receive_response():
            messages.append(message)

    # Verify we got a result
    result_messages = [m for m in messages if isinstance(m, ResultMessage)]
    assert result_messages, f"No ResultMessage. Got: {[type(m).__name__ for m in messages]}"
    assert not result_messages[0].is_error, f"Query error: {result_messages[0].errors}"

    # Inspect all content blocks from assistant and user messages
    all_content_blocks: list[dict[str, Any]] = []
    for msg in messages:
        if isinstance(msg, AssistantMessage):
            for block in msg.content:
                all_content_blocks.append(
                    {"source": "assistant", "type": block.type, "block": block}
                )
        elif isinstance(msg, UserMessage) and isinstance(msg.content, list):
            for block in msg.content:
                all_content_blocks.append({"source": "user", "type": block.type, "block": block})

    # We expect at least a tool_use block (Claude calling the tool)
    tool_use_blocks = [b for b in all_content_blocks if b["type"] == "tool_use"]
    assert tool_use_blocks, "No tool_use blocks found - Claude didn't call the MCP tool"

    # Verify the tool_use targeted our MCP tool
    tool_use_block = tool_use_blocks[0]["block"]
    assert isinstance(tool_use_block, ToolUseBlock)
    assert tool_use_block.name == "mcp__image_test__get_test_image"

    # Find tool_result blocks and verify image content via get_parsed_content()
    tool_result_blocks = [b for b in all_content_blocks if b["type"] == "tool_result"]
    assert tool_result_blocks, "No tool_result blocks found"

    result_block = tool_result_blocks[0]["block"]
    assert isinstance(result_block, ToolResultBlock)
    assert isinstance(result_block.content, list), "Expected list content in tool result"

    # Parse into typed Anthropic SDK content blocks
    parsed = result_block.get_parsed_content()
    assert isinstance(parsed, list)
    assert len(parsed) >= 1

    # Find the image block in parsed content
    image_params = [b for b in parsed if isinstance(b, dict) and b.get("type") == "image"]
    assert image_params, f"No image block in parsed content: {[type(b).__name__ for b in parsed]}"

    image_param = image_params[0]
    # BetaImageBlockParam is a TypedDict with source.type, source.data, source.media_type
    assert image_param["type"] == "image"
    assert "source" in image_param
    assert image_param["source"]["type"] == "base64"
    assert image_param["source"]["media_type"] == "image/png"
    assert len(image_param["source"]["data"]) > 0, "Image data should not be empty"


@pytest.mark.e2e
@pytest.mark.asyncio
async def test_mcp_progress_tool_wire_format():
    """Test that MCP tool progress notifications flow through the wire protocol.

    Configures Claude Code with a FastMCP stdio server that has a tool
    reporting progress via ctx.report_progress(), then asks Claude to call it
    and verifies that ToolProgressMessage events are received and the tool
    completes successfully.
    """
    mcp_server_path = str(Path(__file__).parent.parent / "mcp_server.py")

    options = ClaudeAgentOptions(
        mcp_servers={
            "progress_test": {
                "type": "stdio",
                "command": sys.executable,
                "args": [mcp_server_path],
            },
        },
        permission_mode="bypassPermissions",
        allow_dangerously_skip_permissions=True,
        max_turns=3,
    )

    messages: list[Message] = []
    async with ClaudeSDKClient(options=options) as client:
        await client.query(
            'Call the mcp__progress_test__test_progress tool with message "hello from test".'
        )
        async for message in client.receive_response():
            messages.append(message)

    # Verify we got a successful result
    result_messages = [m for m in messages if isinstance(m, ResultMessage)]
    assert result_messages, f"No ResultMessage. Got: {[type(m).__name__ for m in messages]}"
    assert not result_messages[0].is_error, f"Query error: {result_messages[0].errors}"

    # Collect assistant content blocks to verify tool_use
    all_content_blocks: list[dict[str, Any]] = []
    for msg in messages:
        if isinstance(msg, AssistantMessage):
            for block in msg.content:
                all_content_blocks.append(
                    {"source": "assistant", "type": block.type, "block": block}
                )
        elif isinstance(msg, UserMessage) and isinstance(msg.content, list):
            for block in msg.content:
                all_content_blocks.append({"source": "user", "type": block.type, "block": block})

    # Verify Claude called the progress tool
    tool_use_blocks = [b for b in all_content_blocks if b["type"] == "tool_use"]
    assert tool_use_blocks, "No tool_use blocks found - Claude didn't call the MCP tool"

    tool_use_block = tool_use_blocks[0]["block"]
    assert isinstance(tool_use_block, ToolUseBlock)
    assert tool_use_block.name == "mcp__progress_test__test_progress"

    # Verify the tool result contains the expected completion message
    tool_result_blocks = [b for b in all_content_blocks if b["type"] == "tool_result"]
    assert tool_result_blocks, "No tool_result blocks found"

    result_block = tool_result_blocks[0]["block"]
    assert isinstance(result_block, ToolResultBlock)
    parsed = result_block.get_parsed_content()
    # The result should contain the completion message as text
    if isinstance(parsed, str):
        assert "hello from test" in parsed
    elif isinstance(parsed, list):
        text_parts = [b.text for b in parsed if isinstance(b, TextBlock)]  # type: ignore[union-attr]
        combined = " ".join(text_parts)
        assert "hello from test" in combined, f"Expected message in result, got: {combined}"

    progress_messages = [m for m in messages if isinstance(m, ToolProgressMessage)]
    # Right now no progress messages emitted from the subprocess, seems to be a bug.
    # rever assertion in case it changes
    assert not progress_messages
    for pm in progress_messages:
        assert pm.tool_name, "Progress message should have a tool_name"
        assert pm.elapsed_time_seconds >= 0, "elapsed_time_seconds should be non-negative"


if __name__ == "__main__":
    pytest.main([__file__, "-vv", "-m", "e2e"])



"""Claude Code storage models.

Pydantic models for Claude Code's JSONL transcript format, enabling
interoperability between agentpool and Claude Code.

Key features:
- JSONL-based conversation logs per project
- Multi-agent support (main + sub-agents)
- Message ancestry tracking
- Conversation forking and branching
- Discriminated unions for type-safe parsing

See ARCHITECTURE.md for detailed documentation of the storage format.
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import Annotated, Any, Literal, assert_never

from pydantic import BaseModel, Discriminator, Field, Tag

from clawd_code_sdk.models.base import ClaudeCodeBaseModel, StopReason
from clawd_code_sdk.models.output_types import ToolUseResult  # noqa: TC001


# See https://github.com/daaain/claude-code-log/blob/main/claude_code_log/models.py

UserType = Literal["external", "internal"]
MCPToolCallStatus = Literal["started", "completed", "failed"]


class ClaudeTextBlock(BaseModel):
    """Text content block."""

    type: Literal["text"]
    text: str


class ClaudeToolUseBlock(BaseModel):
    """Tool use content block."""

    type: Literal["tool_use"]
    id: str
    name: str
    input: dict[str, Any]


class ClaudeToolResultBlock(BaseModel):
    """Tool result content block."""

    type: Literal["tool_result"]
    tool_use_id: str
    content: list[dict[str, Any]] | str | None = None
    is_error: bool | None = None

    def extract_text(self) -> str:
        """Extract text content from this tool result."""
        match self.content:
            case None:
                return ""
            case str():
                return self.content
            case list():
                text_parts = [
                    tc.get("text", "")
                    for tc in self.content
                    if isinstance(tc, dict) and tc.get("type") == "text"
                ]
                return "\n".join(text_parts)
            case _ as unreachable:
                assert_never(unreachable)


class ClaudeThinkingBlock(BaseModel):
    """Thinking/reasoning content block."""

    type: Literal["thinking"]
    thinking: str
    signature: str | None = None


class ClaudeImageSource(BaseModel):
    """Base64-encoded image source data."""

    type: Literal["base64"]
    media_type: str
    data: str


class ClaudeImageBlock(BaseModel):
    """Image content block."""

    type: Literal["image"]
    source: ClaudeImageSource


ClaudeContentBlock = Annotated[
    ClaudeTextBlock
    | ClaudeToolUseBlock
    | ClaudeToolResultBlock
    | ClaudeThinkingBlock
    | ClaudeImageBlock,
    Field(discriminator="type"),
]
"""Discriminated union of all content block types in message content arrays."""


# =============================================================================
# Legacy flat content model (for backward compatibility)
# =============================================================================


class ClaudeMessageContent(BaseModel):
    """Content block in Claude message (flat model).

    This is kept for backward compatibility with code that accesses
    content blocks via a single model with optional fields.
    Prefer using ClaudeContentBlock for new code.
    """

    type: Literal["text", "tool_use", "tool_result", "thinking", "image"]
    # For text blocks
    text: str | None = None
    # For tool_use blocks
    id: str | None = None
    name: str | None = None
    input: dict[str, Any] | None = None
    # For tool_result blocks
    tool_use_id: str | None = None
    content: list[dict[str, Any]] | str | None = None
    is_error: bool | None = None
    # For thinking blocks
    thinking: str | None = None
    signature: str | None = None
    # For image blocks
    source: ClaudeImageSource | None = None

    def extract_tool_result_content(self) -> str:
        """Extract content from a tool_result block."""
        match self.content:
            case None:
                return ""
            case str():
                return self.content
            case list():
                text_parts = [
                    tc.get("text", "")
                    for tc in self.content
                    if isinstance(tc, dict) and tc.get("type") == "text"
                ]
                return "\n".join(text_parts)
            case _ as unreachable:
                assert_never(unreachable)


# =============================================================================
# Token usage
# =============================================================================


class ClaudeUsage(BaseModel):
    """Token usage from Claude API response."""

    input_tokens: int = 0
    output_tokens: int = 0
    cache_creation_input_tokens: int = 0
    cache_read_input_tokens: int = 0
    service_tier: str | None = None
    server_tool_use: dict[str, Any] | None = None

    def add(self, other: ClaudeUsage) -> None:
        self.input_tokens += other.input_tokens
        self.output_tokens += other.output_tokens
        self.cache_creation_input_tokens += other.cache_creation_input_tokens
        self.cache_read_input_tokens += other.cache_read_input_tokens


# =============================================================================
# Message models (role-based)
# =============================================================================


class ClaudeApiMessage(BaseModel):
    """Claude API assistant message structure."""

    model: str
    id: str
    type: Literal["message"] = "message"
    role: Literal["assistant"]
    content: str | Sequence[ClaudeContentBlock]
    stop_reason: StopReason | None = None
    stop_sequence: str | None = None
    usage: ClaudeUsage = Field(default_factory=ClaudeUsage)


class ClaudeUserMessage(BaseModel):
    """User message content."""

    role: Literal["user"]
    content: str | Sequence[ClaudeContentBlock]
    usage: ClaudeUsage | None = None
    """Usage info (for type compatibility with ClaudeApiMessage)."""


# =============================================================================
# JSONL entry base
# =============================================================================


class ClaudeEntryBase(ClaudeCodeBaseModel):
    """Common fields shared across user/assistant/system/saved_hook_context entries."""

    uuid: str
    parent_uuid: str | None = None
    session_id: str
    timestamp: str

    # Context
    cwd: str = ""
    git_branch: str | None = None
    version: str = ""

    # Metadata
    user_type: UserType = "external"
    is_sidechain: bool = False
    is_meta: bool | None = None
    agent_id: str | None = None


# =============================================================================
# User / Assistant entries
# =============================================================================


class ClaudeMessageEntryBase(ClaudeEntryBase):
    """Base for user/assistant message entries with message payload."""

    message: ClaudeApiMessage | ClaudeUserMessage

    is_compact_summary: bool | None = None
    request_id: str | None = None


class ClaudeUserEntry(ClaudeMessageEntryBase):
    """User message entry."""

    type: Literal["user"]
    tool_use_result: (
        list[ToolUseResult | dict[str, Any]] | ToolUseResult | dict[str, Any] | str | None
    ) = None


class ClaudeAssistantEntry(ClaudeMessageEntryBase):
    """Assistant message entry."""

    type: Literal["assistant"]
    is_api_error_message: bool | None = None


ClaudeEntry = ClaudeUserEntry | ClaudeAssistantEntry


# =============================================================================
# Queue operation entries (discriminated by "operation")
# =============================================================================


class ClaudeDocumentContent(ClaudeCodeBaseModel):
    """Document content block."""

    type: Literal["document"]
    source: ClaudeImageSource


ClaudeQueueContent = (
    str | ClaudeTextBlock | ClaudeImageBlock | ClaudeDocumentContent | ClaudeToolResultBlock
)


class ClaudeEnqueueOperation(ClaudeCodeBaseModel):
    """Enqueue operation with content."""

    type: Literal["queue-operation"]
    operation: Literal["enqueue"]
    content: str | list[str | ClaudeQueueContent]
    session_id: str
    timestamp: str


class ClaudeDequeueOperation(ClaudeCodeBaseModel):
    """Dequeue operation."""

    type: Literal["queue-operation"]
    operation: Literal["dequeue"]
    session_id: str
    timestamp: str


class ClaudeRemoveOperation(ClaudeCodeBaseModel):
    """Remove operation (steering input)."""

    type: Literal["queue-operation"]
    operation: Literal["remove"]
    content: str | list[str | ClaudeQueueContent] | None = None
    session_id: str
    timestamp: str


class ClaudePopAllOperation(ClaudeCodeBaseModel):
    """PopAll operation - clears all queued content."""

    type: Literal["queue-operation"]
    operation: Literal["popAll"]
    content: str
    session_id: str
    timestamp: str


ClaudeQueueOperationEntry = Annotated[
    ClaudeEnqueueOperation | ClaudeDequeueOperation | ClaudeRemoveOperation | ClaudePopAllOperation,
    Field(discriminator="operation"),
]


# =============================================================================
# System entries (discriminated by "subtype")
# =============================================================================


class ClaudeSystemEntryBase(ClaudeEntryBase):
    """Common fields shared across all system entry subtypes."""

    type: Literal["system"]
    slug: str | None = None


class ClaudeHookInfo(ClaudeCodeBaseModel):
    """Hook info in stop_hook_summary."""

    command: str


class ClaudeCompactBoundaryEntry(ClaudeSystemEntryBase):
    """System entry emitted when conversation is compacted."""

    subtype: Literal["compact_boundary"]
    content: str | None = None
    level: str | None = None
    logical_parent_uuid: str | None = None
    compact_metadata: dict[str, Any] | None = None


class ClaudeTurnDurationEntry(ClaudeSystemEntryBase):
    """System entry recording the duration of a turn."""

    subtype: Literal["turn_duration"]
    duration_ms: int | None = None


class ClaudeApiErrorEntry(ClaudeSystemEntryBase):
    """System entry for API errors with retry information."""

    subtype: Literal["api_error"]
    level: str | None = None
    error: dict[str, Any] | None = None
    retry_in_ms: float | None = None
    retry_attempt: int | None = None
    max_retries: int | None = None


class ClaudeLocalCommandEntry(ClaudeSystemEntryBase):
    """System entry for local slash commands."""

    subtype: Literal["local_command"]
    content: str | None = None
    level: str | None = None


class ClaudeStopHookSummaryEntry(ClaudeSystemEntryBase):
    """System entry summarizing hook execution at turn end."""

    subtype: Literal["stop_hook_summary"]
    hook_count: int | None = None
    hook_infos: list[ClaudeHookInfo] | None = None
    hook_errors: list[Any] | None = None
    prevented_continuation: bool | None = None
    stop_reason: str | None = None
    has_output: bool | None = None


class ClaudeGenericSystemEntry(ClaudeSystemEntryBase):
    """Fallback for unknown or future system entry subtypes.

    Claude Code keeps adding new system subtypes. This model captures
    any subtype not explicitly modeled, preserving all data as optional fields.
    """

    subtype: str | None = None
    content: str | None = None
    level: str | None = None

    # Fields that may appear on various subtypes
    duration_ms: int | None = None
    is_compact_summary: bool | None = None
    logical_parent_uuid: str | None = None
    compact_metadata: dict[str, Any] | None = None
    hook_count: int | None = None
    hook_infos: list[ClaudeHookInfo] | None = None
    hook_errors: list[Any] | None = None
    prevented_continuation: bool | None = None
    stop_reason: str | None = None
    has_output: bool | None = None
    error: dict[str, Any] | None = None
    retry_in_ms: float | None = None
    retry_attempt: int | None = None
    max_retries: int | None = None
    tool_use_id: str | None = Field(default=None, alias="toolUseID")
    tool_use_result: (
        list[ToolUseResult | dict[str, Any]] | ToolUseResult | dict[str, Any] | str | None
    ) = None


def _system_entry_discriminator(data: Any) -> str:  # noqa: ANN401
    """Discriminator function for system entry subtypes.

    Routes known subtypes to their typed models, falls back to generic.
    Pydantic calls this with a dict (from JSON) or a model instance (re-validation).
    """
    subtype: str | None = None
    match data:
        case dict():
            subtype = data.get("subtype") or data.get("sub_type")
        case (
            ClaudeCompactBoundaryEntry()
            | ClaudeTurnDurationEntry()
            | ClaudeApiErrorEntry()
            | ClaudeLocalCommandEntry()
            | ClaudeStopHookSummaryEntry()
        ) as entry:
            subtype = entry.subtype
        case ClaudeGenericSystemEntry() as entry:
            subtype = entry.subtype
    if subtype in _KNOWN_SYSTEM_SUBTYPES:
        return subtype
    return "__generic__"


_KNOWN_SYSTEM_SUBTYPES: frozenset[str] = frozenset(
    {
        "compact_boundary",
        "turn_duration",
        "api_error",
        "local_command",
        "stop_hook_summary",
    }
)

ClaudeSystemEntry = Annotated[
    Annotated[ClaudeCompactBoundaryEntry, Tag("compact_boundary")]
    | Annotated[ClaudeTurnDurationEntry, Tag("turn_duration")]
    | Annotated[ClaudeApiErrorEntry, Tag("api_error")]
    | Annotated[ClaudeLocalCommandEntry, Tag("local_command")]
    | Annotated[ClaudeStopHookSummaryEntry, Tag("stop_hook_summary")]
    | Annotated[ClaudeGenericSystemEntry, Tag("__generic__")],
    Discriminator(_system_entry_discriminator),
]
"""Discriminated union of system entry subtypes with generic fallback."""


# =============================================================================
# Summary entry
# =============================================================================


class ClaudeSummaryEntry(ClaudeCodeBaseModel):
    """Summary entry (conversation summary)."""

    type: Literal["summary"]
    leaf_uuid: str
    summary: str
    cwd: str | None = None
    session_id: str | None = None
    """Summaries may not have a session_id."""


# =============================================================================
# File history snapshot entry
# =============================================================================


class ClaudeFileHistorySnapshot(ClaudeCodeBaseModel):
    """Snapshot data in file history entry."""

    message_id: str
    tracked_file_backups: dict[str, Any]
    timestamp: str


class ClaudeFileHistoryEntry(ClaudeCodeBaseModel):
    """File history snapshot entry."""

    type: Literal["file-history-snapshot"]
    message_id: str
    snapshot: ClaudeFileHistorySnapshot | dict[str, Any]
    is_snapshot_update: bool = False


# =============================================================================
# Progress entries (data discriminated by "type")
# =============================================================================


class ClaudeMcpProgressData(ClaudeCodeBaseModel):
    """Progress data for MCP tool operations."""

    type: Literal["mcp_progress"]
    status: MCPToolCallStatus | None = None
    server_name: str | None = None
    tool_name: str | None = None
    elapsed_time_ms: int | None = None


class ClaudeBashProgressData(ClaudeCodeBaseModel):
    """Progress data for bash tool operations."""

    type: Literal["bash_progress"]
    output: str | None = None
    full_output: str | None = None
    elapsed_time_seconds: int | None = None
    total_lines: int | None = None


class ClaudeHookProgressData(ClaudeCodeBaseModel):
    """Progress data for hook operations."""

    type: Literal["hook_progress"]
    hook_event: str | None = None
    hook_name: str | None = None
    command: str | None = None
    prompt_text: str | None = None
    status_message: str | None = None


class ClaudeWaitingForTaskData(ClaudeCodeBaseModel):
    """Progress data for waiting task operations."""

    type: Literal["waiting_for_task"]
    task_description: str | None = None
    task_type: str | None = None


class ClaudeAgentProgressData(ClaudeCodeBaseModel):
    """Progress data for agent/subagent operations."""

    type: Literal["agent_progress"]
    prompt: str | None = None
    agent_id: str | None = None
    message: dict[str, Any] | None = None
    normalized_messages: list[dict[str, Any]] | None = None
    resume: dict[str, Any] | None = None


class ClaudeQueryUpdateData(ClaudeCodeBaseModel):
    """Progress data for search query updates."""

    type: Literal["query_update"]
    query: str


class ClaudeSearchResultsReceivedData(ClaudeCodeBaseModel):
    """Progress data for search results received."""

    type: Literal["search_results_received"]
    result_count: int
    query: str


class ClaudeSkillProgressData(ClaudeCodeBaseModel):
    """Progress data for skill operations."""

    type: Literal["skill_progress"]
    prompt: str | None = None
    agent_id: str | None = None


class ClaudeTaskProgressData(ClaudeCodeBaseModel):
    """Progress data for task tracking."""

    type: Literal["task_progress"]
    task_id: str | None = None
    task_type: str | None = None
    message: str | None = None


class ClaudeToolProgressData(ClaudeCodeBaseModel):
    """Progress data for tool execution (SDK format)."""

    type: Literal["tool_progress"]
    tool_use_id: str | None = None
    tool_name: str | None = None
    parent_tool_use_id: str | None = None
    elapsed_time_seconds: float | None = None
    session_id: str | None = None


ClaudeProgressData = Annotated[
    ClaudeMcpProgressData
    | ClaudeBashProgressData
    | ClaudeHookProgressData
    | ClaudeWaitingForTaskData
    | ClaudeAgentProgressData
    | ClaudeQueryUpdateData
    | ClaudeSearchResultsReceivedData
    | ClaudeSkillProgressData
    | ClaudeTaskProgressData
    | ClaudeToolProgressData,
    Field(discriminator="type"),
]


class ClaudeProgressEntry(ClaudeCodeBaseModel):
    """Progress entry for tracking tool execution status."""

    type: Literal["progress"]
    uuid: str
    slug: str | None = None
    parent_uuid: str | None = None
    session_id: str
    timestamp: str
    data: ClaudeProgressData
    tool_use_id: str | None = None
    parent_tool_use_id: str | None = None
    agent_id: str | None = None
    # Common fields
    cwd: str = ""
    git_branch: str = ""
    version: str = ""
    user_type: UserType = "external"
    is_sidechain: bool = False


# =============================================================================
# Saved hook context entry
# =============================================================================


class ClaudeSavedHookContextEntry(ClaudeEntryBase):
    """Saved hook context entry.

    Stores hook context data (e.g. injected system prompts from hooks)
    persisted alongside the conversation.
    """

    type: Literal["saved_hook_context"]
    content: list[str] | str | None = None
    hook_name: str | None = None
    hook_event: str | None = None
    tool_use_id: str | None = Field(default=None, alias="toolUseID")


# =============================================================================
# Top-level discriminated union
# =============================================================================

ClaudeJSONLEntry = Annotated[
    ClaudeUserEntry
    | ClaudeAssistantEntry
    | ClaudeQueueOperationEntry
    | ClaudeSystemEntry
    | ClaudeSummaryEntry
    | ClaudeFileHistoryEntry
    | ClaudeProgressEntry
    | ClaudeSavedHookContextEntry,
    Field(discriminator="type"),
]
"""Discriminated union for all JSONL entry types."""



"""SDK control protocol types."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated, Any, Literal, TypedDict

from pydantic import Discriminator, TypeAdapter

from .agents import AgentDefinition  # noqa: TC001
from .base import PermissionMode  # noqa: TC001
from .hooks import HookEvent, HookInput  # noqa: TC001
from .mcp import ExternalMcpServerConfig, JSONRPCMessage  # noqa: TC001


# SDK Control Protocol
@dataclass(frozen=True, slots=True, kw_only=True)
class SDKControlInterruptRequest:
    subtype: Literal["interrupt"] = "interrupt"


@dataclass(frozen=True, slots=True, kw_only=True)
class SDKControlPermissionRequest:
    subtype: Literal["can_use_tool"] = "can_use_tool"
    tool_name: str
    input: dict[str, Any]
    tool_use_id: str
    permission_suggestions: list[Any] | None = None
    blocked_path: str | None = None
    decision_reason: str | None = None
    agent_id: str | None = None
    description: str | None = None


@dataclass(frozen=True, slots=True, kw_only=True)
class SDKControlInitializeRequest:
    subtype: Literal["initialize"] = "initialize"
    hooks: dict[HookEvent, Any] | None = None
    agents: dict[str, AgentDefinition] | None = None
    sdk_mcp_servers: list[str] | None = None
    system_prompt: str | None = None
    append_system_prompt: str | None = None
    json_schema: dict[str, Any] | None = None
    prompt_suggestions: bool | None = None


@dataclass(frozen=True, slots=True, kw_only=True)
class SDKControlSetPermissionModeRequest:
    subtype: Literal["set_permission_mode"] = "set_permission_mode"
    mode: PermissionMode


@dataclass(frozen=True, slots=True, kw_only=True)
class SDKHookCallbackRequest:
    subtype: Literal["hook_callback"] = "hook_callback"
    callback_id: str
    input: HookInput
    tool_use_id: str | None = None


@dataclass(frozen=True, slots=True, kw_only=True)
class SDKControlMcpMessageRequest:
    subtype: Literal["mcp_message"] = "mcp_message"
    server_name: str
    message: JSONRPCMessage


@dataclass(frozen=True, slots=True, kw_only=True)
class SDKControlRewindFilesRequest:
    subtype: Literal["rewind_files"] = "rewind_files"
    user_message_id: str
    dry_run: bool | None = None


@dataclass(frozen=True, slots=True, kw_only=True)
class SDKControlStopTaskRequest:
    subtype: Literal["stop_task"] = "stop_task"
    task_id: str


@dataclass(frozen=True, slots=True, kw_only=True)
class SDKControlApplyFlagSettingsRequest:
    subtype: Literal["apply_flag_settings"] = "apply_flag_settings"
    settings: dict[str, Any]


@dataclass(frozen=True, slots=True, kw_only=True)
class SDKControlSetModelRequest:
    """Sets the model to use for subsequent conversation turns."""

    subtype: Literal["set_model"] = "set_model"
    model: str | None = None


@dataclass(frozen=True, slots=True, kw_only=True)
class SDKControlSetMaxThinkingTokensRequest:
    """Sets the maximum number of thinking tokens for extended thinking."""

    subtype: Literal["set_max_thinking_tokens"] = "set_max_thinking_tokens"
    max_thinking_tokens: int | None = None


@dataclass(frozen=True, slots=True, kw_only=True)
class SDKControlMcpStatusRequest:
    """Requests the current status of all MCP server connections."""

    subtype: Literal["mcp_status"] = "mcp_status"


@dataclass(frozen=True, slots=True, kw_only=True)
class SDKControlMcpSetServersRequest:
    """Replaces the set of dynamically managed MCP servers."""

    subtype: Literal["mcp_set_servers"] = "mcp_set_servers"
    servers: dict[str, ExternalMcpServerConfig]


@dataclass(frozen=True, slots=True, kw_only=True)
class SDKControlMcpReconnectRequest:
    """Reconnects a disconnected or failed MCP server."""

    subtype: Literal["mcp_reconnect"] = "mcp_reconnect"
    server_name: str


@dataclass(frozen=True, slots=True, kw_only=True)
class SDKControlMcpToggleRequest:
    """Enables or disables an MCP server."""

    subtype: Literal["mcp_toggle"] = "mcp_toggle"
    server_name: str
    enabled: bool


@dataclass(frozen=True, slots=True, kw_only=True)
class SDKControlMcpAuthenticateRequest:
    """Authenticates with an MCP server."""

    subtype: Literal["mcp_authenticate"] = "mcp_authenticate"
    server_name: str


@dataclass(frozen=True, slots=True, kw_only=True)
class SDKControlMcpClearAuthRequest:
    """Clears authentication for an MCP server."""

    subtype: Literal["mcp_clear_auth"] = "mcp_clear_auth"
    server_name: str


ControlRequestUnion = Annotated[
    SDKControlInterruptRequest
    | SDKControlPermissionRequest
    | SDKControlInitializeRequest
    | SDKControlSetPermissionModeRequest
    | SDKControlSetModelRequest
    | SDKControlSetMaxThinkingTokensRequest
    | SDKControlMcpStatusRequest
    | SDKHookCallbackRequest
    | SDKControlMcpMessageRequest
    | SDKControlRewindFilesRequest
    | SDKControlMcpSetServersRequest
    | SDKControlMcpReconnectRequest
    | SDKControlMcpToggleRequest
    | SDKControlMcpAuthenticateRequest
    | SDKControlMcpClearAuthRequest
    | SDKControlStopTaskRequest
    | SDKControlApplyFlagSettingsRequest,
    Discriminator("subtype"),
]

control_request_adapter: TypeAdapter[ControlRequestUnion] = TypeAdapter(ControlRequestUnion)


@dataclass(frozen=True, slots=True)
class SDKControlRequest:
    type: Literal["control_request"]
    request_id: str
    request: ControlRequestUnion


@dataclass(frozen=True, slots=True)
class SDKControlCancelRequest:
    """Cancels a currently open control request."""

    type: Literal["control_cancel_request"]
    request_id: str


class ControlResponse(TypedDict):
    subtype: Literal["success"]
    request_id: str
    response: dict[str, Any] | None


class ControlErrorResponse(TypedDict):
    subtype: Literal["error"]
    request_id: str
    error: str


class SDKControlResponse(TypedDict):
    type: Literal["control_response"]
    response: ControlResponse | ControlErrorResponse



"""Error types for Claude SDK."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal


if TYPE_CHECKING:
    from anthropic.types.model import Model


class ClaudeSDKError(Exception):
    """Base exception for all Claude SDK errors."""


class CLIConnectionError(ClaudeSDKError):
    """Raised when unable to connect to Claude Code."""


class CLINotFoundError(CLIConnectionError):
    """Raised when Claude Code is not found or not installed."""

    def __init__(self, message: str = "Claude Code not found", cli_path: str | None = None):
        if cli_path:
            message = f"{message}: {cli_path}"
        super().__init__(message)


class ProcessError(ClaudeSDKError):
    """Raised when the CLI process fails."""

    def __init__(self, message: str, exit_code: int | None = None, stderr: str | None = None):
        self.exit_code = exit_code
        self.stderr = stderr

        if exit_code is not None:
            message = f"{message} (exit code: {exit_code})"
        if stderr:
            message = f"{message}\nError output: {stderr}"

        super().__init__(message)


class CLIJSONDecodeError(ClaudeSDKError):
    """Raised when unable to decode JSON from CLI output."""

    def __init__(self, line: str, original_error: Exception):
        self.line = line
        self.original_error = original_error
        super().__init__(f"Failed to decode JSON: {line[:100]}...")


class MessageParseError(ClaudeSDKError):
    """Raised when unable to parse a message from CLI output."""

    def __init__(self, message: str, data: dict[str, Any] | None = None):
        self.data = data
        super().__init__(message)


# API Error types - raised when the Anthropic API returns errors
# These correspond to the error types in types.AssistantMessageError
APIErrorType = Literal[
    "authentication_failed",
    "billing_error",
    "rate_limit",
    "invalid_request",
    "server_error",
    "unknown",
]


class APIError(ClaudeSDKError):
    """Base exception for Anthropic API errors.

    This exception is raised when the API returns an error response that was
    previously returned as a text message. Subclasses provide more specific
    error types for programmatic handling.

    Attributes:
        error_type: The type of API error (e.g., "rate_limit", "invalid_request").
        message: The error message from the API.
        model: The model that was being used when the error occurred.
    """

    def __init__(
        self,
        message: str,
        error_type: APIErrorType,
        model: Model | None = None,
    ):
        self.error_type = error_type
        self.model = model
        super().__init__(message)


class AuthenticationError(APIError):
    """Raised when API authentication fails (401).

    This typically indicates an invalid API key or insufficient permissions.
    """

    def __init__(self, message: str, model: Model | None = None):
        super().__init__(message, "authentication_failed", model)


class BillingError(APIError):
    """Raised when there's a billing issue with the API account.

    This may indicate insufficient credits, expired subscription, or other
    billing-related issues.
    """

    def __init__(self, message: str, model: Model | None = None):
        super().__init__(message, "billing_error", model)


class RateLimitError(APIError):
    """Raised when API rate limits are exceeded (429).

    Applications should implement retry logic with exponential backoff
    when handling this exception.
    """

    def __init__(self, message: str, model: Model | None = None):
        super().__init__(message, "rate_limit", model)


class InvalidRequestError(APIError):
    """Raised when the API request is invalid (400).

    This may indicate invalid parameters, unsupported model, or malformed input.
    Check the error message for details about what needs to be corrected.
    """

    def __init__(self, message: str, model: Model | None = None):
        super().__init__(message, "invalid_request", model)


class ServerError(APIError):
    """Raised when the API server encounters an error (500/529).

    This includes internal server errors (500) and overload errors (529).
    Applications should implement retry logic for transient server errors.
    """

    def __init__(self, message: str, model: Model | None = None):
        super().__init__(message, "server_error", model)


class ControlRequestError(ClaudeSDKError):
    """Raised when a control protocol request fails."""

    def __init__(self, message: str, subtype: str | None = None):
        self.subtype = subtype
        super().__init__(message)


class ControlRequestTimeoutError(ControlRequestError):
    """Raised when a control protocol request times out."""


class McpError(ControlRequestError):
    """Raised when an MCP-related control request fails.

    Attributes:
        error_code: The JSON-RPC error code from the MCP server.
        server_name: The name of the MCP server, if known.
    """

    def __init__(
        self,
        message: str,
        *,
        error_code: int | None = None,
        server_name: str | None = None,
        subtype: str | None = None,
    ):
        self.error_code = error_code
        self.server_name = server_name
        super().__init__(message, subtype=subtype)



"""Tests for subprocess transport buffering edge cases."""

from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any
from unittest.mock import AsyncMock, MagicMock

import anyio
import pytest

from clawd_code_sdk._errors import CLIJSONDecodeError
from clawd_code_sdk._internal.transport.subprocess_cli import (
    _DEFAULT_MAX_BUFFER_SIZE,
    SubprocessCLITransport,
)
from clawd_code_sdk.models import ClaudeAgentOptions


if TYPE_CHECKING:
    from collections.abc import AsyncIterator

DEFAULT_CLI_PATH = "/usr/bin/claude"


def make_options(**kwargs: Any) -> ClaudeAgentOptions:
    """Construct ClaudeAgentOptions with a default CLI path for tests."""

    cli_path = kwargs.pop("cli_path", DEFAULT_CLI_PATH)
    assert isinstance(cli_path, str)
    return ClaudeAgentOptions(cli_path=cli_path, **kwargs)


class MockTextReceiveStream:
    """Mock TextReceiveStream for testing."""

    def __init__(self, lines: list[str]) -> None:
        self.lines = lines
        self.index = 0

    def __aiter__(self) -> AsyncIterator[str]:
        return self

    async def __anext__(self) -> str:
        if self.index >= len(self.lines):
            raise StopAsyncIteration
        line = self.lines[self.index]
        self.index += 1
        return line


class TestSubprocessBuffering:
    """Test subprocess transport handling of buffered output."""

    def test_multiple_json_objects_on_single_line(self) -> None:
        """Test parsing when multiple JSON objects are concatenated on a single line.

        In some environments, stdout buffering can cause multiple distinct JSON
        objects to be delivered as a single line with embedded newlines.
        """

        async def _test() -> None:
            json_obj1 = {"type": "message", "id": "msg1", "content": "First message"}
            json_obj2 = {"type": "result", "id": "res1", "status": "completed"}
            buffered_line = json.dumps(json_obj1) + "\n" + json.dumps(json_obj2)
            transport = SubprocessCLITransport(prompt="test", options=make_options())
            mock_process = MagicMock()
            mock_process.returncode = None
            mock_process.wait = AsyncMock(return_value=None)
            transport._process = mock_process
            transport._stdout_stream = MockTextReceiveStream([buffered_line])  # type: ignore[assignment]
            transport._stderr_stream = MockTextReceiveStream([])  # type: ignore[assignment]

            messages: list[Any] = []
            async for msg in transport.read_messages():
                messages.append(msg)

            assert len(messages) == 2
            assert messages[0]["type"] == "message"
            assert messages[0]["id"] == "msg1"
            assert messages[0]["content"] == "First message"
            assert messages[1]["type"] == "result"
            assert messages[1]["id"] == "res1"
            assert messages[1]["status"] == "completed"

        anyio.run(_test)

    def test_json_with_embedded_newlines(self) -> None:
        """Test parsing JSON objects that contain newline characters in string values."""

        async def _test() -> None:
            json_obj1 = {"type": "message", "content": "Line 1\nLine 2\nLine 3"}
            json_obj2 = {"type": "result", "data": "Some\nMultiline\nContent"}
            buffered_line = json.dumps(json_obj1) + "\n" + json.dumps(json_obj2)
            transport = SubprocessCLITransport(prompt="test", options=make_options())
            mock_process = MagicMock()
            mock_process.returncode = None
            mock_process.wait = AsyncMock(return_value=None)
            transport._process = mock_process
            transport._stdout_stream = MockTextReceiveStream([buffered_line])  # pyright: ignore[reportAttributeAccessIssue]
            transport._stderr_stream = MockTextReceiveStream([])  # pyright: ignore[reportAttributeAccessIssue]
            messages: list[Any] = []
            async for msg in transport.read_messages():
                messages.append(msg)

            assert len(messages) == 2
            assert messages[0]["content"] == "Line 1\nLine 2\nLine 3"
            assert messages[1]["data"] == "Some\nMultiline\nContent"

        anyio.run(_test)

    def test_multiple_newlines_between_objects(self) -> None:
        """Test parsing with multiple newlines between JSON objects."""

        async def _test() -> None:
            json_obj1 = {"type": "message", "id": "msg1"}
            json_obj2 = {"type": "result", "id": "res1"}
            buffered_line = json.dumps(json_obj1) + "\n\n\n" + json.dumps(json_obj2)
            transport = SubprocessCLITransport(prompt="test", options=make_options())
            mock_process = MagicMock()
            mock_process.returncode = None
            mock_process.wait = AsyncMock(return_value=None)
            transport._process = mock_process
            transport._stdout_stream = MockTextReceiveStream([buffered_line])  # pyright: ignore[reportAttributeAccessIssue]
            transport._stderr_stream = MockTextReceiveStream([])  # pyright: ignore[reportAttributeAccessIssue]
            messages: list[dict[str, Any]] = []
            async for msg in transport.read_messages():
                messages.append(msg)

            assert len(messages) == 2
            assert messages[0]["id"] == "msg1"
            assert messages[1]["id"] == "res1"

        anyio.run(_test)

    def test_split_json_across_multiple_reads(self) -> None:
        """Test parsing when a single JSON object is split across multiple stream reads."""

        async def _test() -> None:
            json_obj = {
                "type": "assistant",
                "message": {
                    "content": [
                        {"type": "text", "text": "x" * 1000},
                        {
                            "type": "tool_use",
                            "id": "tool_123",
                            "name": "Read",
                            "input": {"file_path": "/test.txt"},
                        },
                    ]
                },
            }

            complete_json = json.dumps(json_obj)
            part1 = complete_json[:100]
            part2 = complete_json[100:250]
            part3 = complete_json[250:]
            transport = SubprocessCLITransport(prompt="test", options=make_options())
            mock_process = MagicMock()
            mock_process.returncode = None
            mock_process.wait = AsyncMock(return_value=None)
            transport._process = mock_process
            transport._stdout_stream = MockTextReceiveStream([part1, part2, part3])  # pyright: ignore[reportAttributeAccessIssue]
            transport._stderr_stream = MockTextReceiveStream([])  # pyright: ignore[reportAttributeAccessIssue]

            messages: list[Any] = []
            async for msg in transport.read_messages():
                messages.append(msg)

            assert len(messages) == 1
            assert messages[0]["type"] == "assistant"
            assert len(messages[0]["message"]["content"]) == 2

        anyio.run(_test)

    def test_large_minified_json(self) -> None:
        """Test parsing a large minified JSON (simulating the reported issue)."""

        async def _test() -> None:
            large_data = {"data": [{"id": i, "value": "x" * 100} for i in range(1000)]}
            json_obj = {
                "type": "user",
                "message": {
                    "role": "user",
                    "content": [
                        {
                            "tool_use_id": "toolu_016fed1NhiaMLqnEvrj5NUaj",
                            "type": "tool_result",
                            "content": json.dumps(large_data),
                        }
                    ],
                },
            }
            complete_json = json.dumps(json_obj)
            chunk_size = 64 * 1024
            chunks = [
                complete_json[i : i + chunk_size] for i in range(0, len(complete_json), chunk_size)
            ]
            transport = SubprocessCLITransport(prompt="test", options=make_options())
            mock_process = MagicMock()
            mock_process.returncode = None
            mock_process.wait = AsyncMock(return_value=None)
            transport._process = mock_process
            transport._stdout_stream = MockTextReceiveStream(chunks)  # pyright: ignore[reportAttributeAccessIssue]
            transport._stderr_stream = MockTextReceiveStream([])  # pyright: ignore[reportAttributeAccessIssue]

            messages: list[Any] = []
            async for msg in transport.read_messages():
                messages.append(msg)

            assert len(messages) == 1
            assert messages[0]["type"] == "user"
            assert (
                messages[0]["message"]["content"][0]["tool_use_id"]
                == "toolu_016fed1NhiaMLqnEvrj5NUaj"
            )

        anyio.run(_test)

    def test_buffer_size_exceeded(self) -> None:
        """Test that exceeding buffer size raises an appropriate error."""

        async def _test() -> None:
            huge_incomplete = '{"data": "' + "x" * (_DEFAULT_MAX_BUFFER_SIZE + 1000)
            transport = SubprocessCLITransport(prompt="test", options=make_options())
            mock_process = MagicMock()
            mock_process.returncode = None
            mock_process.wait = AsyncMock(return_value=None)
            transport._process = mock_process
            transport._stdout_stream = MockTextReceiveStream([huge_incomplete])  # pyright: ignore[reportAttributeAccessIssue]
            transport._stderr_stream = MockTextReceiveStream([])  # pyright: ignore[reportAttributeAccessIssue]
            with pytest.raises(Exception) as exc_info:
                messages: list[Any] = []
                async for msg in transport.read_messages():
                    messages.append(msg)

            assert isinstance(exc_info.value, CLIJSONDecodeError)
            assert "exceeded maximum buffer size" in str(exc_info.value)

        anyio.run(_test)

    def test_buffer_size_option(self) -> None:
        """Test that the configurable buffer size option is respected."""

        async def _test() -> None:
            custom_limit = 512
            huge_incomplete = '{"data": "' + "x" * (custom_limit + 10)
            opts = make_options(max_buffer_size=custom_limit)
            transport = SubprocessCLITransport(prompt="test", options=opts)
            mock_process = MagicMock()
            mock_process.returncode = None
            mock_process.wait = AsyncMock(return_value=None)
            transport._process = mock_process
            transport._stdout_stream = MockTextReceiveStream([huge_incomplete])  # pyright: ignore[reportAttributeAccessIssue]
            transport._stderr_stream = MockTextReceiveStream([])  # pyright: ignore[reportAttributeAccessIssue]

            with pytest.raises(CLIJSONDecodeError) as exc_info:
                async for _ in transport.read_messages():
                    pass

            assert f"maximum buffer size of {custom_limit} bytes" in str(exc_info.value)

        anyio.run(_test)

    def test_mixed_complete_and_split_json(self) -> None:
        """Test handling a mix of complete and split JSON messages."""

        async def _test() -> None:
            msg1 = json.dumps({"type": "system", "subtype": "start"})
            large_msg = {
                "type": "assistant",
                "message": {"content": [{"type": "text", "text": "y" * 5000}]},
            }
            large_json = json.dumps(large_msg)
            msg3 = json.dumps({"type": "system", "subtype": "end"})
            lines = [
                msg1 + "\n",
                large_json[:1000],
                large_json[1000:3000],
                large_json[3000:] + "\n" + msg3,
            ]
            transport = SubprocessCLITransport(prompt="test", options=make_options())
            mock_process = MagicMock()
            mock_process.returncode = None
            mock_process.wait = AsyncMock(return_value=None)
            transport._process = mock_process
            transport._stdout_stream = MockTextReceiveStream(lines)  # pyright: ignore[reportAttributeAccessIssue]
            transport._stderr_stream = MockTextReceiveStream([])  # pyright: ignore[reportAttributeAccessIssue]
            messages: list[Any] = []
            async for msg in transport.read_messages():
                messages.append(msg)

            assert len(messages) == 3
            assert messages[0]["type"] == "system"
            assert messages[0]["subtype"] == "start"
            assert messages[1]["type"] == "assistant"
            assert len(messages[1]["message"]["content"][0]["text"]) == 5000
            assert messages[2]["type"] == "system"
            assert messages[2]["subtype"] == "end"

        anyio.run(_test)


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



"""Hook system types: inputs, outputs, matchers, and callbacks."""

from __future__ import annotations

from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal, NotRequired, TypedDict


if TYPE_CHECKING:
    from collections.abc import Sequence


HookEvent = Literal[
    "PreToolUse",
    "PostToolUse",
    "PostToolUseFailure",
    "UserPromptSubmit",
    "Stop",
    "SubagentStop",
    "PreCompact",
    "Notification",
    "SubagentStart",
    "PermissionRequest",
    "SessionStart",
    "SessionEnd",
    "Setup",
    "TeammateIdle",
    "TaskCompleted",
    "ConfigChange",
    "WorktreeCreate",
    "WorktreeRemove",
]


# Hook input types - strongly typed for each hook event
class BaseHookInput(TypedDict):
    """Base hook input fields present across many hook events."""

    session_id: str
    transcript_path: str
    cwd: str
    permission_mode: NotRequired[str]


class PreToolUseHookInput(BaseHookInput):
    """Input data for PreToolUse hook events."""

    hook_event_name: Literal["PreToolUse"]
    tool_name: str
    tool_input: dict[str, Any]
    tool_use_id: str


class PostToolUseHookInput(BaseHookInput):
    """Input data for PostToolUse hook events."""

    hook_event_name: Literal["PostToolUse"]
    tool_name: str
    tool_input: dict[str, Any]
    tool_response: Any
    tool_use_id: str


class PostToolUseFailureHookInput(BaseHookInput):
    """Input data for PostToolUseFailure hook events."""

    hook_event_name: Literal["PostToolUseFailure"]
    tool_name: str
    tool_input: dict[str, Any]
    tool_use_id: str
    error: str
    is_interrupt: NotRequired[bool]


class UserPromptSubmitHookInput(BaseHookInput):
    """Input data for UserPromptSubmit hook events."""

    hook_event_name: Literal["UserPromptSubmit"]
    prompt: str


class StopHookInput(BaseHookInput):
    """Input data for Stop hook events."""

    hook_event_name: Literal["Stop"]
    stop_hook_active: bool


class SubagentStopHookInput(BaseHookInput):
    """Input data for SubagentStop hook events."""

    hook_event_name: Literal["SubagentStop"]
    stop_hook_active: bool
    agent_id: str
    agent_transcript_path: str
    agent_type: str


class PreCompactHookInput(BaseHookInput):
    """Input data for PreCompact hook events."""

    hook_event_name: Literal["PreCompact"]
    trigger: Literal["manual", "auto"]
    custom_instructions: str | None


class NotificationHookInput(BaseHookInput):
    """Input data for Notification hook events."""

    hook_event_name: Literal["Notification"]
    message: str
    title: NotRequired[str]
    notification_type: str


class SubagentStartHookInput(BaseHookInput):
    """Input data for SubagentStart hook events."""

    hook_event_name: Literal["SubagentStart"]
    agent_id: str
    agent_type: str


class PermissionRequestHookInput(BaseHookInput):
    """Input data for PermissionRequest hook events."""

    hook_event_name: Literal["PermissionRequest"]
    tool_name: str
    tool_input: dict[str, Any]
    permission_suggestions: NotRequired[list[Any]]


class SessionStartHookInput(BaseHookInput):
    """Input data for SessionStart hook events."""

    hook_event_name: Literal["SessionStart"]
    source: Literal["startup", "resume", "clear", "compact"]
    agent_type: NotRequired[str]
    model: NotRequired[str]


class SessionEndHookInput(BaseHookInput):
    """Input data for SessionEnd hook events."""

    hook_event_name: Literal["SessionEnd"]
    reason: Literal["clear", "logout", "prompt_input_exit", "other", "bypass_permissions_disabled"]


class SetupHookInput(BaseHookInput):
    """Input data for Setup hook events."""

    hook_event_name: Literal["Setup"]
    trigger: Literal["init", "maintenance"]


class TeammateIdleHookInput(BaseHookInput):
    """Input data for TeammateIdle hook events."""

    hook_event_name: Literal["TeammateIdle"]
    teammate_name: str
    team_name: str


class TaskCompletedHookInput(BaseHookInput):
    """Input data for TaskCompleted hook events."""

    hook_event_name: Literal["TaskCompleted"]
    task_id: str
    task_subject: str
    task_description: NotRequired[str]
    teammate_name: NotRequired[str]
    team_name: NotRequired[str]


class ConfigChangeHookInput(BaseHookInput):
    """Input data for ConfigChange hook events."""

    hook_event_name: Literal["ConfigChange"]
    source: Literal[
        "user_settings", "project_settings", "local_settings", "policy_settings", "skills"
    ]
    file_path: NotRequired[str]


class WorktreeCreateHookInput(BaseHookInput):
    """Input data for WorktreeCreate hook events."""

    hook_event_name: Literal["WorktreeCreate"]
    name: str


class WorktreeRemoveHookInput(BaseHookInput):
    """Input data for WorktreeRemove hook events."""

    hook_event_name: Literal["WorktreeRemove"]
    worktree_path: str


# Union type for all hook inputs
HookInput = (
    PreToolUseHookInput
    | PostToolUseHookInput
    | PostToolUseFailureHookInput
    | UserPromptSubmitHookInput
    | StopHookInput
    | SubagentStopHookInput
    | PreCompactHookInput
    | NotificationHookInput
    | SubagentStartHookInput
    | PermissionRequestHookInput
    | SessionStartHookInput
    | SessionEndHookInput
    | SetupHookInput
    | TeammateIdleHookInput
    | TaskCompletedHookInput
    | ConfigChangeHookInput
    | WorktreeCreateHookInput
    | WorktreeRemoveHookInput
)


# Hook-specific output types
class PreToolUseHookSpecificOutput(TypedDict):
    """Hook-specific output for PreToolUse events."""

    hookEventName: Literal["PreToolUse"]
    permissionDecision: NotRequired[Literal["allow", "deny", "ask"]]
    permissionDecisionReason: NotRequired[str]
    updatedInput: NotRequired[dict[str, Any]]
    additionalContext: NotRequired[str]


class PostToolUseHookSpecificOutput(TypedDict):
    """Hook-specific output for PostToolUse events."""

    hookEventName: Literal["PostToolUse"]
    additionalContext: NotRequired[str]
    updatedMCPToolOutput: NotRequired[Any]


class PostToolUseFailureHookSpecificOutput(TypedDict):
    """Hook-specific output for PostToolUseFailure events."""

    hookEventName: Literal["PostToolUseFailure"]
    additionalContext: NotRequired[str]


class UserPromptSubmitHookSpecificOutput(TypedDict):
    """Hook-specific output for UserPromptSubmit events."""

    hookEventName: Literal["UserPromptSubmit"]
    additionalContext: NotRequired[str]


class SessionStartHookSpecificOutput(TypedDict):
    """Hook-specific output for SessionStart events."""

    hookEventName: Literal["SessionStart"]
    additionalContext: NotRequired[str]


class NotificationHookSpecificOutput(TypedDict):
    """Hook-specific output for Notification events."""

    hookEventName: Literal["Notification"]
    additionalContext: NotRequired[str]


class SubagentStartHookSpecificOutput(TypedDict):
    """Hook-specific output for SubagentStart events."""

    hookEventName: Literal["SubagentStart"]
    additionalContext: NotRequired[str]


class PermissionRequestHookSpecificOutput(TypedDict):
    """Hook-specific output for PermissionRequest events."""

    hookEventName: Literal["PermissionRequest"]
    decision: dict[str, Any]


HookSpecificOutput = (
    PreToolUseHookSpecificOutput
    | PostToolUseHookSpecificOutput
    | PostToolUseFailureHookSpecificOutput
    | UserPromptSubmitHookSpecificOutput
    | SessionStartHookSpecificOutput
    | NotificationHookSpecificOutput
    | SubagentStartHookSpecificOutput
    | PermissionRequestHookSpecificOutput
)


# See https://docs.anthropic.com/en/docs/claude-code/hooks#advanced%3A-json-output
# for documentation of the output types.
#
# IMPORTANT: The Python SDK uses `async_` and `continue_` (with underscores) to avoid
# Python keyword conflicts. These fields are automatically converted to `async` and
# `continue` when sent to the CLI. You should use the underscore versions in your
# Python code.
class AsyncHookJSONOutput(TypedDict):
    """Async hook output that defers hook execution.

    Fields:
        async_: Set to True to defer hook execution. Note: This is converted to
            "async" when sent to the CLI - use "async_" in your Python code.
        asyncTimeout: Optional timeout in milliseconds for the async operation.
    """

    async_: Literal[True]  # Using async_ to avoid Python keyword (converted to "async" for CLI)
    asyncTimeout: NotRequired[int]


class SyncHookJSONOutput(TypedDict):
    """Synchronous hook output with control and decision fields.

    This defines the structure for hook callbacks to control execution and provide
    feedback to Claude.

    Common Control Fields:
        continue_: Whether Claude should proceed after hook execution (default: True).
            Note: This is converted to "continue" when sent to the CLI.
        suppressOutput: Hide stdout from transcript mode (default: False).
        stopReason: Message shown when continue is False.

    Decision Fields:
        decision: Set to "block" to indicate blocking behavior.
        systemMessage: Warning message displayed to the user.
        reason: Feedback message for Claude about the decision.

    Hook-Specific Output:
        hookSpecificOutput: Event-specific controls (e.g., permissionDecision for
            PreToolUse, additionalContext for PostToolUse).

    Note: The CLI documentation shows field names without underscores ("async", "continue"),
    but Python code should use the underscore versions ("async_", "continue_") as they
    are automatically converted.
    """

    # Common control fields
    continue_: NotRequired[bool]  # Avoid name clash (converted to "continue" for CLI)
    suppressOutput: NotRequired[bool]
    stopReason: NotRequired[str]
    # Decision fields
    # Note: "approve" is deprecated for PreToolUse (use permissionDecision instead)
    # For other hooks, only "block" is meaningful
    decision: NotRequired[Literal["block"]]
    systemMessage: NotRequired[str]
    reason: NotRequired[str]
    # Hook-specific outputs
    hookSpecificOutput: NotRequired[HookSpecificOutput]


HookJSONOutput = AsyncHookJSONOutput | SyncHookJSONOutput


class HookContext(TypedDict):
    """Context information for hook callbacks.

    Fields:
        signal: Reserved for future abort signal support. Currently always None.
    """

    signal: Any | None  # Future: abort signal support


HookCallback = Callable[
    # HookCallback input parameters:
    # - input: Strongly-typed hook input with discriminated unions based on hook_event_name
    # - tool_use_id: Optional tool use identifier
    # - context: Hook context with abort signal support (currently placeholder)
    [HookInput, str | None, HookContext],
    Awaitable[HookJSONOutput],
]


# Hook matcher configuration
@dataclass
class HookMatcher:
    """Hook matcher configuration."""

    # See https://docs.anthropic.com/en/docs/claude-code/hooks#structure for the
    # expected string value. For example, for PreToolUse, the matcher can be
    # a tool name like "Bash" or a combination of tool names like
    # "Write|MultiEdit|Edit".
    matcher: str | None = None
    # A list of Python functions with function signature HookCallback
    hooks: Sequence[HookCallback] = field(default_factory=list)
    # Timeout in seconds for all hooks in this matcher (default: 60)
    timeout: float | None = None



"""End-to-end test for subagent invocation via AgentDefinition.

Verifies that when the main agent dispatches work to a subagent defined
via AgentDefinition, the SDK emits the expected task lifecycle messages
(task_started, task_notification) and that the subagent actually runs.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from clawd_code_sdk import (
    AgentDefinition,
    AssistantMessage,
    ClaudeAgentOptions,
    ClaudeSDKClient,
    InitSystemMessage,
    ResultMessage,
)
from clawd_code_sdk.models.messages import TaskNotificationSystemMessage, TaskStartedSystemMessage


if TYPE_CHECKING:
    from clawd_code_sdk import Message


@pytest.mark.e2e
@pytest.mark.asyncio
async def test_subagent_task_lifecycle():
    """Test that invoking a subagent produces task_started and task_notification messages.

    Defines a simple subagent, asks the main agent to call it, and verifies:
    1. The subagent appears in the init message
    2. A TaskStartedSystemMessage is emitted when the subagent begins
    3. A TaskNotificationSystemMessage is emitted when the subagent completes
    4. The task_id is consistent between started and notification
    5. The main agent's response includes output from the subagent
    """
    options = ClaudeAgentOptions(
        agents={
            "echo-agent": AgentDefinition(
                description="A simple echo agent that repeats what you say",
                prompt=(
                    "You are an echo agent. When given a message, respond with "
                    "exactly: 'ECHO: ' followed by the message. Nothing else."
                ),
            )
        },
        max_turns=10,
        permission_mode="bypassPermissions",
    )

    messages: list[Message] = []
    async with ClaudeSDKClient(options=options) as client:
        await client.query(
            "Use the echo-agent subagent with the message 'pineapple'. "
            "Include the subagent's response in your reply."
        )
        async for msg in client.receive_response():
            messages.append(msg)

    # 1. Verify agent was registered
    init_msgs = [m for m in messages if isinstance(m, InitSystemMessage)]
    assert init_msgs, "Should have received an InitSystemMessage"
    assert "echo-agent" in init_msgs[0].agents

    # 2. Check for task lifecycle messages
    task_started = [m for m in messages if isinstance(m, TaskStartedSystemMessage)]
    task_notifications = [m for m in messages if isinstance(m, TaskNotificationSystemMessage)]

    assert task_started, (
        "Should have received at least one TaskStartedSystemMessage "
        f"(got message types: {[type(m).__name__ for m in messages]})"
    )
    assert task_notifications, (
        "Should have received at least one TaskNotificationSystemMessage "
        f"(got message types: {[type(m).__name__ for m in messages]})"
    )

    # 3. Verify task_id consistency
    started_task_ids = {m.task_id for m in task_started}
    notified_task_ids = {m.task_id for m in task_notifications}
    assert started_task_ids & notified_task_ids, (
        f"task_id should match between started ({started_task_ids}) "
        f"and notification ({notified_task_ids})"
    )

    # 4. Verify notification status
    completed = [m for m in task_notifications if m.status == "completed"]
    assert completed, (
        f"At least one task should have completed, got statuses: "
        f"{[m.status for m in task_notifications]}"
    )

    # 5. Verify the main agent's response includes subagent output
    response_text = ""
    for msg in messages:
        if isinstance(msg, AssistantMessage) and not hasattr(msg, "parent_tool_use_id"):
            for block in msg.content:
                if hasattr(block, "text"):
                    response_text += block.text

    response_lower = response_text.lower()
    assert "pineapple" in response_lower, (
        f"Response should include 'pineapple' from the echo agent, got: {response_text[:500]}"
    )

    # 6. Verify we got a final result
    result_msgs = [m for m in messages if isinstance(m, ResultMessage)]
    assert result_msgs, "Should have received a ResultMessage"



"""Anthropic SDK types for tool result content blocks.

This module defines a discriminated union of all possible content types
that can appear in tool results from the Anthropic SDK. These types are
used to provide proper typing for ToolResultBlock.content.
"""

from __future__ import annotations

from typing import Annotated

from anthropic.types import TextBlock, WebSearchResultBlock, WebSearchToolResultError
from anthropic.types.beta import (
    BetaBashCodeExecutionResultBlock,
    BetaBashCodeExecutionToolResultError,
    BetaCodeExecutionResultBlock,
    BetaCodeExecutionToolResultError,
    BetaImageBlockParam,
    BetaTextEditorCodeExecutionCreateResultBlock,
    BetaTextEditorCodeExecutionStrReplaceResultBlock,
    BetaTextEditorCodeExecutionToolResultError,
    BetaTextEditorCodeExecutionViewResultBlock,
    BetaToolReferenceBlock,
    BetaToolSearchToolResultError,
    BetaToolSearchToolSearchResultBlock,
    BetaWebFetchBlock,
    BetaWebFetchToolResultErrorBlock,
)
from pydantic import BaseModel, Field, TypeAdapter


# Union of all possible content types that can appear in tool results.
# These are the inner content blocks, not the outer tool result wrapper.
# Discriminated by the "type" field.
#
# Type discriminator values:
# - "text" -> TextBlock
# - "image" -> BetaImageBlockParam (TypedDict, no Pydantic model exists)
# - "tool_reference" -> BetaToolReferenceBlock
# - "tool_search_tool_search_result" -> BetaToolSearchToolSearchResultBlock
# - "tool_search_tool_result_error" -> BetaToolSearchToolResultError
# - "web_search_result" -> WebSearchResultBlock
# - "web_search_tool_result_error" -> WebSearchToolResultError
# - "web_fetch_result" -> BetaWebFetchBlock
# - "web_fetch_tool_result_error" -> BetaWebFetchToolResultErrorBlock
# - "code_execution_result" -> BetaCodeExecutionResultBlock
# - "code_execution_tool_result_error" -> BetaCodeExecutionToolResultError
# - "bash_code_execution_result" -> BetaBashCodeExecutionResultBlock
# - "bash_code_execution_tool_result_error" -> BetaBashCodeExecutionToolResultError
# - "text_editor_code_execution_view_result" -> BetaTextEditorCodeExecutionViewResultBlock
# - "text_editor_code_execution_create_result" -> BetaTextEditorCodeExecutionCreateResultBlock
# - "text_editor_code_execution_str_replace_result" -> BetaTextEditorCodeExecutionStrReplaceResultBlock
# - "text_editor_code_execution_tool_result_error" -> BetaTextEditorCodeExecutionToolResultError
ToolResultContentBlock = Annotated[
    TextBlock
    | BetaImageBlockParam
    | BetaToolReferenceBlock
    | BetaToolSearchToolSearchResultBlock
    | BetaToolSearchToolResultError
    | WebSearchResultBlock
    | WebSearchToolResultError
    | BetaWebFetchBlock
    | BetaWebFetchToolResultErrorBlock
    | BetaCodeExecutionResultBlock
    | BetaCodeExecutionToolResultError
    | BetaBashCodeExecutionResultBlock
    | BetaBashCodeExecutionToolResultError
    | BetaTextEditorCodeExecutionViewResultBlock
    | BetaTextEditorCodeExecutionCreateResultBlock
    | BetaTextEditorCodeExecutionStrReplaceResultBlock
    | BetaTextEditorCodeExecutionToolResultError,
    Field(discriminator="type"),
]
_tool_result_content_adapter: TypeAdapter[list[ToolResultContentBlock]] | None = None


def _get_adapter() -> TypeAdapter[list[ToolResultContentBlock]]:
    global _tool_result_content_adapter
    if _tool_result_content_adapter is None:
        # Force schema build for Anthropic models (deferred by default)
        for model in [
            TextBlock,
            WebSearchResultBlock,
            WebSearchToolResultError,
            BetaBashCodeExecutionResultBlock,
            BetaBashCodeExecutionToolResultError,
            BetaCodeExecutionResultBlock,
            BetaCodeExecutionToolResultError,
            BetaTextEditorCodeExecutionCreateResultBlock,
            BetaTextEditorCodeExecutionStrReplaceResultBlock,
            BetaTextEditorCodeExecutionToolResultError,
            BetaTextEditorCodeExecutionViewResultBlock,
            BetaToolReferenceBlock,
            BetaToolSearchToolResultError,
            BetaToolSearchToolSearchResultBlock,
            BetaWebFetchBlock,
            BetaWebFetchToolResultErrorBlock,
        ]:
            if isinstance(model, type) and issubclass(model, BaseModel):
                model.model_rebuild()
        _tool_result_content_adapter = TypeAdapter(list[ToolResultContentBlock])
    return _tool_result_content_adapter


def validate_tool_result_content(content: list[dict[str, object]]) -> list[ToolResultContentBlock]:
    """Validate and parse raw tool result content into typed blocks.

    Args:
        content: Raw list of content block dictionaries from CLI output

    Returns:
        List of validated and typed content blocks
    """
    return _get_adapter().validate_python(content)
